@@ -1,9 +1,5 @@
-# -*- coding: utf-8 -*-
-
 """sievelib."""
 
-from __future__ import unicode_literals
-
 from pkg_resources import get_distribution, DistributionNotFound
 
 
