@@ -1,5 +1,3 @@
-# coding: utf-8
-
 """
 SIEVE commands representation
 
@@ -21,12 +19,9 @@ provides extra information such as:
  * etc.
 
 """
-from __future__ import unicode_literals
-
-from collections import Iterable
 import sys
-
-from future.utils import python_2_unicode_compatible
+from collections.abc import Iterable
+from . import tools
 
 
 class CommandError(Exception):
@@ -35,7 +30,6 @@ class CommandError(Exception):
     pass
 
 
-@python_2_unicode_compatible
 class UnknownCommand(CommandError):
     """Specific exception raised when an unknown command is encountered"""
 
@@ -46,7 +40,6 @@ class UnknownCommand(CommandError):
         return "unknown command %s" % self.name
 
 
-@python_2_unicode_compatible
 class BadArgument(CommandError):
     """Specific exception raised when a bad argument is encountered"""
 
@@ -60,7 +53,6 @@ class BadArgument(CommandError):
                % (self.seen, self.command, self.expected)
 
 
-@python_2_unicode_compatible
 class BadValue(CommandError):
     """Specific exception raised when a bad argument value is encountered"""
 
@@ -73,7 +65,6 @@ class BadValue(CommandError):
                % (self.value, self.argument)
 
 
-@python_2_unicode_compatible
 class ExtensionNotLoaded(CommandError):
     """Raised when an extension is not loaded."""
 
@@ -83,6 +74,7 @@ class ExtensionNotLoaded(CommandError):
     def __str__(self):
         return "extension '{}' not loaded".format(self.name)
 
+
 # Statement elements (see RFC, section 8.3)
 # They are used in different commands.
 comparator = {"name": "comparator",
@@ -95,10 +87,21 @@ address_part = {"name": "address-part",
                 "values": [":localpart", ":domain", ":all"],
                 "type": ["tag"],
                 "required": False}
-match_type = {"name": "match-type",
-              "values": [":is", ":contains", ":matches"],
-              "type": ["tag"],
-              "required": False}
+match_type = {
+    "name": "match-type",
+    "values": [":is", ":contains", ":matches"],
+    "extension_values": {
+        ":count": "relational",
+        ":value": "relational"
+    },
+    "extra_arg": {
+        "type": "string",
+        "values": ['"gt"', '"ge"', '"lt"', '"le"', '"eq"', '"ne"'],
+        "valid_for": [":count", ":value"]
+    },
+    "type": ["tag"],
+    "required": False,
+}
 
 
 class Command(object):
@@ -116,13 +119,15 @@ class Command(object):
     """
     _type = None
     variable_args_nb = False
+    non_deterministic_args = False
     accept_children = False
     must_follow = None
-    is_extension = False
+    extension = None
 
     def __init__(self, parent=None):
         self.parent = parent
         self.arguments = {}
+        self.extra_arguments = {}  # to store tag arguments
         self.children = []
 
         self.nextargpos = 0
@@ -153,9 +158,15 @@ class Command(object):
                     continue
                 target.write(" ")
                 value = self.arguments[arg["name"]]
-
-                if "tag" in arg["type"] and arg.get("write_tag", False):
-                    target.write("%s " % arg["values"][0])
+                atype = arg["type"]
+                if "tag" in atype:
+                    target.write(value)
+                    if arg["name"] in self.extra_arguments:
+                        value = self.extra_arguments[arg["name"]]
+                        atype = arg["extra_arg"]["type"]
+                        target.write(" ")
+                    else:
+                        continue
 
                 if type(value) == list:
                     if self.__get_arg_type(arg["name"]) == ["testlist"]:
@@ -176,9 +187,9 @@ class Command(object):
                     value.tosieve(indentlevel, target=target)
                     continue
 
-                if "string" in arg["type"]:
+                if "string" in atype:
                     target.write(value)
-                    if not value.startswith('"'):
+                    if not value.startswith('"') and not value.startswith("["):
                         target.write("\n")
                 else:
                     target.write(value)
@@ -225,6 +236,14 @@ class Command(object):
     def has_arguments(self):
         return len(self.args_definition) != 0
 
+    def reassign_arguments(self):
+        """Reassign arguments to proper slots.
+
+        Should be called when parsing of commands with non
+        deterministic arguments is considered done.
+        """
+        raise NotImplementedError
+
     def dump(self, indentlevel=0, target=sys.stdout):
         """Display the command
 
@@ -240,6 +259,14 @@ class Command(object):
                 if not arg["name"] in self.arguments:
                     continue
                 value = self.arguments[arg["name"]]
+                atype = arg["type"]
+                if "tag" in atype:
+                    self.__print(str(value), indentlevel, target=target)
+                    if arg["name"] in self.extra_arguments:
+                        value = self.extra_arguments[arg["name"]]
+                        atype = arg["extra_arg"]["type"]
+                    else:
+                        continue
                 if type(value) == list:
                     if self.__get_arg_type(arg["name"]) == ["testlist"]:
                         for t in value:
@@ -255,6 +282,26 @@ class Command(object):
         for ch in self.children:
             ch.dump(indentlevel, target)
 
+    def walk(self):
+        """Walk through commands."""
+        yield self
+        if self.has_arguments():
+            for arg in self.args_definition:
+                if not arg["name"] in self.arguments:
+                    continue
+                value = self.arguments[arg["name"]]
+                if type(value) == list:
+                    if self.__get_arg_type(arg["name"]) == ["testlist"]:
+                        for t in value:
+                            for node in t.walk():
+                                yield node
+                if isinstance(value, Command):
+                    for node in value.walk():
+                        yield node
+        for ch in self.children:
+            for node in ch.walk():
+                yield node
+
     def addchild(self, child):
         """Add a new child to the command
 
@@ -269,7 +316,7 @@ class Command(object):
         self.children += [child]
         return True
 
-    def iscomplete(self):
+    def iscomplete(self, atype=None, avalue=None):
         """Check if the command is complete
 
         Check if all required arguments have been encountered. For
@@ -283,10 +330,16 @@ class Command(object):
         if self.required_args == -1:
             self.required_args = 0
             for arg in self.args_definition:
-                if arg["required"]:
+                if arg.get("required", False):
                     self.required_args += 1
-        return (not self.curarg or "extra_arg" not in self.curarg) \
-            and (self.rargs_cnt == self.required_args)
+        return (
+            (not self.curarg or
+             "extra_arg" not in self.curarg or
+             ("valid_for" in self.curarg["extra_arg"] and
+              atype and atype in self.curarg["extra_arg"]["type"] and
+              avalue not in self.curarg["extra_arg"]["valid_for"])) and
+            (self.rargs_cnt == self.required_args)
+        )
 
     def get_type(self):
         """Return the command's type"""
@@ -294,7 +347,7 @@ class Command(object):
             raise NotImplementedError
         return self._type
 
-    def __is_valid_value_for_arg(self, arg, value):
+    def __is_valid_value_for_arg(self, arg, value, check_extension=True):
         """Check if value is allowed for arg
 
         Some commands only allow a limited set of values. The method
@@ -303,11 +356,37 @@ class Command(object):
 
         :param arg: the argument's name
         :param value: the value to check
+        :param check_extension: check if value requires an extension
         :return: True on succes, False otherwise
         """
-        if "values" not in arg:
+        if "values" not in arg and "extension_values" not in arg:
             return True
-        return value.lower() in arg["values"]
+        if "values" in arg and value.lower() in arg["values"]:
+            return True
+        if "extension_values" in arg:
+            extension = arg["extension_values"].get(value.lower())
+            if extension:
+                condition = (
+                    check_extension and
+                    extension not in RequireCommand.loaded_extensions
+                )
+                if condition:
+                    raise ExtensionNotLoaded(extension)
+                return True
+        return False
+
+    def __is_valid_type(self, typ, typlist):
+        """ Check if type is valid based on input type list
+            "string" is special because it can be used for stringlist
+
+        :param typ: the type to check
+        :param typlist: the list of type to check
+        :return: True on success, False otherwise
+        """
+        typ_is_str = typ == "string"
+        str_list_in_typlist = "stringlist" in typlist
+
+        return typ in typlist or (typ_is_str and str_list_in_typlist)
 
     def check_next_arg(self, atype, avalue, add=True, check_extension=True):
         """Argument validity checking
@@ -339,24 +418,27 @@ class Command(object):
         """
         if not self.has_arguments():
             return False
-        if self.iscomplete():
+        if self.iscomplete(atype, avalue):
             return False
 
         if self.curarg is not None and "extra_arg" in self.curarg:
-            if atype in self.curarg["extra_arg"]["type"]:
-                if "values" not in self.curarg["extra_arg"] \
-                   or avalue in self.curarg["extra_arg"]["values"]:
-                    if add:
-                        self.arguments[self.curarg["name"]] = avalue
-                    self.curarg = None
-                    return True
+            condition = (
+                atype in self.curarg["extra_arg"]["type"] and
+                ("values" not in self.curarg["extra_arg"] or
+                 avalue in self.curarg["extra_arg"]["values"])
+            )
+            if condition:
+                if add:
+                    self.extra_arguments[self.curarg["name"]] = avalue
+                self.curarg = None
+                return True
             raise BadValue(self.curarg["name"], avalue)
 
         failed = False
         pos = self.nextargpos
         while pos < len(self.args_definition):
             curarg = self.args_definition[pos]
-            if curarg["required"]:
+            if curarg.get("required", False):
                 if curarg["type"] == ["testlist"]:
                     if atype != "test":
                         failed = True
@@ -364,8 +446,9 @@ class Command(object):
                         if not curarg["name"] in self.arguments:
                             self.arguments[curarg["name"]] = []
                         self.arguments[curarg["name"]] += [avalue]
-                elif atype not in curarg["type"] or \
-                        not self.__is_valid_value_for_arg(curarg, avalue):
+                elif not self.__is_valid_type(atype, curarg["type"]) or \
+                        not self.__is_valid_value_for_arg(
+                            curarg, avalue, check_extension):
                     failed = True
                 else:
                     self.curarg = curarg
@@ -375,20 +458,27 @@ class Command(object):
                         self.arguments[curarg["name"]] = avalue
                 break
 
-            if atype in curarg["type"]:
+            condition = (
+                atype in curarg["type"] and
+                self.__is_valid_value_for_arg(curarg, avalue, check_extension)
+            )
+            if condition:
                 ext = curarg.get("extension")
                 condition = (
                     check_extension and ext and
                     ext not in RequireCommand.loaded_extensions)
                 if condition:
                     raise ExtensionNotLoaded(ext)
-                if self.__is_valid_value_for_arg(curarg, avalue):
-                    if "extra_arg" in curarg:
-                        self.curarg = curarg
-                        break
-                    if add:
-                        self.arguments[curarg["name"]] = avalue
-                    break
+                condition = (
+                    "extra_arg" in curarg and
+                    ("valid_for" not in curarg["extra_arg"] or
+                     avalue in curarg["extra_arg"]["valid_for"])
+                )
+                if condition:
+                    self.curarg = curarg
+                if add:
+                    self.arguments[curarg["name"]] = avalue
+                break
 
             pos += 1
 
@@ -449,10 +539,6 @@ class RequireCommand(ControlCommand):
                 RequireCommand.loaded_extensions += [ext]
 
 
-class StopCommand(ControlCommand):
-    args_definition = []
-
-
 class IfCommand(ControlCommand):
     accept_children = True
 
@@ -487,17 +573,54 @@ class ElseCommand(ControlCommand):
 
 class ActionCommand(Command):
     """Indermediate class to represent "action" commands"""
+
     _type = "action"
 
+    def args_as_tuple(self):
+        args = []
+        for name, value in list(self.arguments.items()):
+            unquote = False
+            for argdef in self.args_definition:
+                if name == argdef["name"]:
+                    condition = (
+                        "string" in argdef["type"] or
+                        "stringlist" in argdef["type"]
+                    )
+                    if condition:
+                        unquote = True
+                        break
+            if unquote:
+                if "," in value:
+                    args += tools.to_list(value)
+                else:
+                    args.append(value.strip('"'))
+                continue
+            args.append(value)
+        return (self.name, ) + tuple(args)
+
+
+class StopCommand(ActionCommand):
+    args_definition = []
+
 
 class FileintoCommand(ActionCommand):
-    is_extension = True
+    extension = "fileinto"
     args_definition = [
         {"name": "copy",
          "type": ["tag"],
          "values": [":copy"],
          "required": False,
          "extension": "copy"},
+        {"name": "create",
+         "type": ["tag"],
+         "values": [":create"],
+         "required": False,
+         "extension": "mailbox"},
+        {"name": "flags",
+         "type": ["tag"],
+         "values": [":flags"],
+         "extra_arg": {"type": ["string", "stringlist"]},
+         "extension": "imap4flags"},
         {"name": "mailbox",
          "type": ["string"],
          "required": True}
@@ -518,7 +641,7 @@ class RedirectCommand(ActionCommand):
 
 
 class RejectCommand(ActionCommand):
-    is_extension = True
+    extension = "reject"
     args_definition = [
         {"name": "text",
          "type": ["string"],
@@ -527,13 +650,60 @@ class RejectCommand(ActionCommand):
 
 
 class KeepCommand(ActionCommand):
-    args_definition = []
+    args_definition = [
+        {"name": "flags",
+         "type": ["tag"],
+         "values": [":flags"],
+         "extra_arg": {"type": ["string", "stringlist"]},
+         "extension": "imap4flags"},
+    ]
 
 
 class DiscardCommand(ActionCommand):
     args_definition = []
 
 
+class SetflagCommand(ActionCommand):
+    """imap4flags extension: setflag."""
+
+    args_definition = [
+        {"name": "variable-name",
+         "type": ["string"],
+         "required": False},
+        {"name": "list-of-flags",
+         "type": ["string", "stringlist"],
+         "required": True}
+    ]
+    extension = "imap4flags"
+
+
+class AddflagCommand(ActionCommand):
+    """imap4flags extension: addflag."""
+
+    args_definition = [
+        {"name": "variable-name",
+         "type": ["string"],
+         "required": False},
+        {"name": "list-of-flags",
+         "type": ["string", "stringlist"],
+         "required": True}
+    ]
+    extension = "imap4flags"
+
+
+class RemoveflagCommand(ActionCommand):
+    """imap4flags extension: removeflag."""
+
+    args_definition = [
+        {"name": "variable-name",
+         "type": ["string"]},
+        {"name": "list-of-flags",
+         "type": ["string", "stringlist"],
+         "required": True}
+    ]
+    extension = "imap4flags"
+
+
 class TestCommand(Command):
     """Indermediate class to represent "test" commands"""
     _type = "test"
@@ -593,15 +763,54 @@ class EnvelopeCommand(TestCommand):
          "type": ["string", "stringlist"],
          "required": True}
     ]
+    extension = "envelope"
+
+    def args_as_tuple(self):
+        """Return arguments as a list."""
+        result = ("envelope", self.arguments["match-type"])
+        value = self.arguments["header-list"]
+        if isinstance(value, list):
+            # FIXME
+            value = "[{}]".format(
+                ",".join('"{}"'.format(item) for item in value))
+        if value.startswith("["):
+            result += (tools.to_list(value),)
+        else:
+            result += ([value.strip('"')],)
+        value = self.arguments["key-list"]
+        if isinstance(value, list):
+            # FIXME
+            value = "[{}]".format(
+                ",".join('"{}"'.format(item) for item in value))
+        if value.startswith("["):
+            result += (tools.to_list(value), )
+        else:
+            result = result + ([value.strip('"')],)
+        return result
 
 
 class ExistsCommand(TestCommand):
     args_definition = [
         {"name": "header-names",
-         "type": ["stringlist"],
+         "type": ["string", "stringlist"],
          "required": True}
     ]
 
+    def args_as_tuple(self):
+        """FIXME: en fonction de la manière dont la commande a été générée
+        (factory ou parser), le type des arguments est différent :
+        string quand ça vient de la factory ou type normal depuis le
+        parser. Il faut uniformiser tout ça !!
+
+        """
+        value = self.arguments["header-names"]
+        if isinstance(value, list):
+            value = "[{}]".format(
+                ",".join('"{}"'.format(item) for item in value))
+        if not value.startswith("["):
+            return ('exists', value.strip('"'))
+        return ("exists", ) + tuple(tools.to_list(value))
+
 
 class TrueCommand(TestCommand):
     args_definition = []
@@ -623,6 +832,57 @@ class HeaderCommand(TestCommand):
          "required": True}
     ]
 
+    def args_as_tuple(self):
+        """Return arguments as a list."""
+        if "," in self.arguments["header-names"]:
+            result = tuple(tools.to_list(self.arguments["header-names"]))
+        else:
+            result = (self.arguments["header-names"].strip('"'),)
+        result = result + (self.arguments["match-type"],)
+        if "," in self.arguments["key-list"]:
+            result = result + tuple(
+                tools.to_list(self.arguments["key-list"], unquote=False))
+        else:
+            result = result + (self.arguments["key-list"].strip('"'),)
+        return result
+
+
+class BodyCommand(TestCommand):
+    """Body extension.
+
+    See https://tools.ietf.org/html/rfc5173.
+    """
+
+    args_definition = [
+        comparator,
+        match_type,
+        {"name": "body-transform",
+         "values": [":raw", ":content", ":text"],
+         "extra_arg": {"type": "stringlist", "valid_for": [":content"]},
+         "type": ["tag"],
+         "required": False},
+        {"name": "key-list",
+         "type": ["string", "stringlist"],
+         "required": True},
+    ]
+    extension = "body"
+
+    def args_as_tuple(self):
+        """Return arguments as a list."""
+        result = ("body", )
+        result = result + (
+            self.arguments["body-transform"], self.arguments["match-type"])
+        value = self.arguments["key-list"]
+        if isinstance(value, list):
+            # FIXME
+            value = "[{}]".format(
+                ",".join('"{}"'.format(item) for item in value))
+        if value.startswith("["):
+            result += tuple(tools.to_list(value))
+        else:
+            result += (value.strip('"'),)
+        return result
+
 
 class NotCommand(TestCommand):
     accept_children = True
@@ -648,42 +908,140 @@ class SizeCommand(TestCommand):
          "required": True},
     ]
 
+    def args_as_tuple(self):
+        return ("size", self.arguments["comparator"], self.arguments["limit"])
+
+
+class HasflagCommand(TestCommand):
+    """imap4flags extension: hasflag."""
+
+    args_definition = [
+        comparator,
+        match_type,
+        {"name": "variable-list",
+         "type": ["string", "stringlist"],
+         "required": False},
+        {"name": "list-of-flags",
+         "type": ["string", "stringlist"],
+         "required": True}
+    ]
+    extension = "imap4flags"
+    non_deterministic_args = True
+
+    def reassign_arguments(self):
+        """Deal with optional stringlist before a required one."""
+        condition = (
+            "variable-list" in self.arguments and
+            "list-of-flags" not in self.arguments
+        )
+        if condition:
+            self.arguments["list-of-flags"] = (
+                self.arguments.pop("variable-list"))
+            self.rargs_cnt = 1
+
+
+class DateCommand(TestCommand):
+    """date command, part of the date extension.
+
+    https://tools.ietf.org/html/rfc5260#section-4
+    """
+
+    extension = "date"
+    args_definition = [
+        {"name": "zone",
+         "type": ["tag"],
+         "values": [":zone", ":originalzone"],
+         "extra_arg": {"type": "string", "valid_for": [":zone"]},
+         "required": False},
+        comparator,
+        match_type,
+        {"name": "header-name",
+         "type": ["string"],
+         "required": True},
+        {"name": "date-part",
+         "type": ["string"],
+         "required": True},
+        {"name": "key-list",
+         "type": ["string", "stringlist"],
+         "required": True}
+    ]
+
+
+class CurrentdateCommand(TestCommand):
+    """currentdate command, part of the date extension.
+
+    http://tools.ietf.org/html/rfc5260#section-5
+    """
+
+    extension = "date"
+    args_definition = [
+        {"name": "zone",
+         "type": ["tag"],
+         "values": [":zone"],
+         "extra_arg": {"type": "string"},
+         "required": False},
+        comparator,
+        match_type,
+        {"name": "date-part",
+         "type": ["string"],
+         "required": True},
+        {"name": "key-list",
+         "type": ["string", "stringlist"],
+         "required": True}
+    ]
+
+    def args_as_tuple(self):
+        """Return arguments as a list."""
+        result = ("currentdate", )
+        result += (
+            ":zone",
+            self.extra_arguments["zone"].strip('"'),
+            self.arguments["match-type"],
+        )
+        if self.arguments["match-type"] in [":count", ":value"]:
+            result += (self.extra_arguments["match-type"].strip('"'), )
+        result += (self.arguments["date-part"].strip('"'), )
+        value = self.arguments["key-list"]
+        if isinstance(value, list):
+            # FIXME
+            value = "[{}]".format(
+                ",".join('"{}"'.format(item) for item in value))
+        if value.startswith("["):
+            result = result + tuple(tools.to_list(value))
+        else:
+            result = result + (value.strip('"'),)
+        return result
+
 
 class VacationCommand(ActionCommand):
     args_definition = [
         {"name": "subject",
          "type": ["tag"],
-         "write_tag": True,
          "values": [":subject"],
          "extra_arg": {"type": "string"},
          "required": False},
         {"name": "days",
          "type": ["tag"],
-         "write_tag": True,
          "values": [":days"],
          "extra_arg": {"type": "number"},
          "required": False},
         {"name": "from",
          "type": ["tag"],
-         "write_tag": True,
          "values": [":from"],
          "extra_arg": {"type": "string"},
          "required": False},
         {"name": "addresses",
          "type": ["tag"],
-         "write_tag": True,
          "values": [":addresses"],
          "extra_arg": {"type": ["string", "stringlist"]},
          "required": False},
         {"name": "handle",
          "type": ["tag"],
-         "write_tag": True,
          "values": [":handle"],
          "extra_arg": {"type": "string"},
          "required": False},
         {"name": "mime",
          "type": ["tag"],
-         "write_tag": True,
          "values": [":mime"],
          "required": False},
         {"name": "reason",
@@ -694,12 +1052,12 @@ class VacationCommand(ActionCommand):
 
 class SetCommand(ControlCommand):
 
-    """currentdate command, part of the variables extension
+    """set command, part of the variables extension
 
     http://tools.ietf.org/html/rfc5229
     """
 
-    is_extension = True
+    extension = "variables"
     args_definition = [
         {"name": "startend",
          "type": ["string"],
@@ -710,37 +1068,6 @@ class SetCommand(ControlCommand):
     ]
 
 
-class CurrentdateCommand(ControlCommand):
-
-    """currentdate command, part of the date extension
-
-    http://tools.ietf.org/html/rfc5260#section-5
-    """
-
-    is_extension = True
-    accept_children = True
-    args_definition = [
-        {"name": "zone",
-         "type": ["tag"],
-         "write_tag": True,
-         "values": [":zone"],
-         "extra_arg": {"type": "string"},
-         "required": False},
-        {"name": "match-value",
-         "type": ["tag"],
-         "required": True},
-        {"name": "comparison",
-         "type": ["string"],
-         "required": True},
-        {"name": "match-against",
-         "type": ["string"],
-         "required": True},
-        {"name": "match-against-field",
-         "type": ["string"],
-         "required": True}
-    ]
-
-
 def add_commands(cmds):
     """
     Adds one or more commands to the module namespace.
@@ -771,21 +1098,13 @@ def get_command_instance(name, parent=None, checkexists=True):
     :param parent: the eventual parent command
     :return: a new class instance
     """
-
-    # Mapping between extension names and command names
-    extension_map = {
-        'date': set(['currentdate']),
-        'variables': set(['set'])
-    }
-    extname = name
-    for extension in extension_map:
-        if name in extension_map[extension]:
-            extname = extension
-            break
-
     cname = "%sCommand" % name.lower().capitalize()
-    if cname not in globals() or \
-            (checkexists and globals()[cname].is_extension and
-             extname not in RequireCommand.loaded_extensions):
+    gl = globals()
+    condition = (
+        cname not in gl or
+        (checkexists and gl[cname].extension and
+         gl[cname].extension not in RequireCommand.loaded_extensions)
+    )
+    if condition:
         raise UnknownCommand(name)
-    return globals()[cname](parent)
+    return gl[cname](parent)
