@@ -1,5 +1,3 @@
-# coding: utf-8
-
 """
 Tools for simpler sieve filters generation.
 
@@ -10,19 +8,12 @@ Only commands (control/test/action) defined in the ``commands`` module
 are supported.
 """
 
-from __future__ import print_function, unicode_literals
-
+import io
 import sys
 
-from future.utils import python_2_unicode_compatible
-import six
-
-from sievelib.commands import (
-    get_command_instance, IfCommand, RequireCommand, FalseCommand
-)
+from sievelib import commands
 
 
-@python_2_unicode_compatible
 class FiltersSet(object):
 
     """A set of filters."""
@@ -44,7 +35,7 @@ class FiltersSet(object):
         self.filters = []
 
     def __str__(self):
-        target = six.StringIO()
+        target = io.StringIO()
         self.tosieve(target)
         ret = target.getvalue()
         target.close()
@@ -57,16 +48,16 @@ class FiltersSet(object):
 
         :param fcontent: the filter's name
         """
-        if not isinstance(fcontent, IfCommand):
+        if not isinstance(fcontent, commands.IfCommand):
             return False
-        if not isinstance(fcontent["test"], FalseCommand):
+        if not isinstance(fcontent["test"], commands.FalseCommand):
             return False
         return True
 
     def from_parser_result(self, parser):
         cpt = 1
         for f in parser.result:
-            if isinstance(f, RequireCommand):
+            if isinstance(f, commands.RequireCommand):
                 if type(f.arguments["capabilities"]) == list:
                     [self.require(c) for c in f.arguments["capabilities"]]
                 else:
@@ -76,7 +67,7 @@ class FiltersSet(object):
             name = "Unnamed rule %d" % cpt
             description = ""
             for comment in f.hash_comments:
-                if isinstance(comment, six.binary_type):
+                if isinstance(comment, bytes):
                     comment = comment.decode("utf-8")
                 if comment.startswith(self.filter_name_pretext):
                     name = comment.replace(self.filter_name_pretext, "")
@@ -112,7 +103,7 @@ class FiltersSet(object):
         """
         if not len(self.requires):
             return None
-        reqcmd = get_command_instance("require")
+        reqcmd = commands.get_command_instance("require")
         reqcmd.check_next_arg("stringlist", self.requires)
         return reqcmd
 
@@ -137,7 +128,7 @@ class FiltersSet(object):
         """
         if tag is None:
             tag = condition[1]
-        cmd = get_command_instance("header", parent)
+        cmd = commands.get_command_instance("header", parent)
         cmd.check_next_arg("tag", tag)
         cmd.check_next_arg("string", self.__quote_if_necessary(condition[0]))
         cmd.check_next_arg("string", self.__quote_if_necessary(condition[2]))
@@ -166,8 +157,8 @@ class FiltersSet(object):
         :param actions: the list of actions
         :param matchtype: "anyof" or "allof"
         """
-        ifcontrol = get_command_instance("if")
-        mtypeobj = get_command_instance(matchtype, ifcontrol)
+        ifcontrol = commands.get_command_instance("if")
+        mtypeobj = commands.get_command_instance(matchtype, ifcontrol)
         for c in conditions:
             if c[0].startswith("not"):
                 negate = True
@@ -176,37 +167,94 @@ class FiltersSet(object):
                 negate = False
                 cname = c[0]
             if cname in ("true", "false"):
-                cmd = get_command_instance(c[0], ifcontrol)
+                cmd = commands.get_command_instance(c[0], ifcontrol)
             elif cname == "size":
-                cmd = get_command_instance("size", ifcontrol)
+                cmd = commands.get_command_instance("size", ifcontrol)
                 cmd.check_next_arg("tag", c[1])
                 cmd.check_next_arg("number", c[2])
             elif cname == "exists":
-                cmd = get_command_instance("exists", ifcontrol)
+                cmd = commands.get_command_instance("exists", ifcontrol)
                 cmd.check_next_arg(
                     "stringlist",
                     "[%s]" % (",".join('"%s"' % val for val in c[1:]))
                 )
+            elif cname == "envelope":
+                cmd = commands.get_command_instance(
+                    "envelope", ifcontrol, False)
+                self.require("envelope")
+                if c[1].startswith(":not"):
+                    comp_tag = c[1].replace("not", "")
+                    negate = True
+                else:
+                    comp_tag = c[1]
+                cmd.check_next_arg("tag", comp_tag)
+                cmd.check_next_arg(
+                    "stringlist",
+                    "[{}]".format(",".join('"{}"'.format(val) for val in c[2]))
+                )
+                cmd.check_next_arg(
+                    "stringlist",
+                    "[{}]".format(",".join('"{}"'.format(val) for val in c[3]))
+                )
+            elif cname == "body":
+                cmd = commands.get_command_instance("body", ifcontrol, False)
+                self.require(cmd.extension)
+                cmd.check_next_arg("tag", c[1])
+                if c[2].startswith(":not"):
+                    comp_tag = c[2].replace("not", "")
+                    negate = True
+                else:
+                    comp_tag = c[2]
+                cmd.check_next_arg("tag", comp_tag)
+                cmd.check_next_arg(
+                    "stringlist",
+                    "[%s]" % (",".join('"%s"' % val for val in c[3:]))
+                )
+            elif cname == "currentdate":
+                cmd = commands.get_command_instance(
+                    "currentdate", ifcontrol, False)
+                self.require(cmd.extension)
+                cmd.check_next_arg("tag", c[1])
+                cmd.check_next_arg("string", self.__quote_if_necessary(c[2]))
+                if c[3].startswith(":not"):
+                    comp_tag = c[3].replace("not", "")
+                    negate = True
+                else:
+                    comp_tag = c[3]
+                cmd.check_next_arg("tag", comp_tag, check_extension=False)
+                next_arg_pos = 4
+                if comp_tag == ":value":
+                    self.require("relational")
+                    cmd.check_next_arg(
+                        "string", self.__quote_if_necessary(c[next_arg_pos]))
+                    next_arg_pos += 1
+                cmd.check_next_arg(
+                    "string", self.__quote_if_necessary(c[next_arg_pos]))
+                next_arg_pos += 1
+                cmd.check_next_arg(
+                    "stringlist",
+                    "[%s]" %
+                    (",".join('"%s"' % val for val in c[next_arg_pos:]))
+                )
             else:
+                # header command fallback
                 if c[1].startswith(':not'):
                     cmd = self.__build_condition(
                         c, ifcontrol, c[1].replace("not", "", 1))
-                    not_cmd = get_command_instance("not", ifcontrol)
-                    not_cmd.check_next_arg("test", cmd)
-                    cmd = not_cmd
+                    negate = True
                 else:
                     cmd = self.__build_condition(c, ifcontrol)
             if negate:
-                not_cmd = get_command_instance("not", ifcontrol)
+                not_cmd = commands.get_command_instance("not", ifcontrol)
                 not_cmd.check_next_arg("test", cmd)
                 cmd = not_cmd
             mtypeobj.check_next_arg("test", cmd)
         ifcontrol.check_next_arg("test", mtypeobj)
 
         for actdef in actions:
-            action = get_command_instance(actdef[0], ifcontrol, False)
-            if action.is_extension:
-                self.require(actdef[0])
+            action = commands.get_command_instance(actdef[0], ifcontrol, False)
+            if action.extension is not None:
+                self.require(action.extension)
             for arg in actdef[1:]:
                 self.check_if_arg_is_extension(arg)
                 if arg.startswith(":"):
@@ -221,7 +269,7 @@ class FiltersSet(object):
     def _unicode_filter_name(self, name):
         """Convert name to unicode if necessary."""
         return (
-            name.decode("utf-8") if isinstance(name, six.binary_type) else name
+            name.decode("utf-8") if isinstance(name, bytes) else name
         )
 
     def addfilter(self, name, conditions, actions, matchtype="anyof"):
@@ -304,6 +352,74 @@ class FiltersSet(object):
                 return f["content"]
         return None
 
+    def get_filter_matchtype(self, name):
+        """Retrieve matchtype of the given filter."""
+        flt = self.getfilter(name)
+        if not flt:
+            return None
+        for node in flt.walk():
+            if isinstance(node, (commands.AllofCommand,
+                                 commands.AnyofCommand)):
+                return node.__class__.__name__.lower().replace("command", "")
+        return None
+
+    def get_filter_conditions(self, name):
+        """Retrieve conditions of the given filter."""
+        flt = self.getfilter(name)
+        if not flt:
+            return None
+        conditions = []
+        negate = False
+        for node in flt.walk():
+            if isinstance(node, commands.NotCommand):
+                negate = True
+            elif isinstance(node, (commands.HeaderCommand,
+                                   commands.SizeCommand,
+                                   commands.ExistsCommand,
+                                   commands.BodyCommand,
+                                   commands.EnvelopeCommand,
+                                   commands.CurrentdateCommand)):
+                args = node.args_as_tuple()
+                if negate:
+                    if node.name in ["header", "envelope"]:
+                        nargs = (
+                            args[0],
+                            ":not{}".format(args[1][1:])
+                        )
+                        if len(args) > 3:
+                            nargs += (args[2:])
+                        else:
+                            nargs += (args[2],)
+                        args = nargs
+                    elif node.name == "body":
+                        args = (
+                            args[:2] +
+                            (":not{}".format(args[2][1:]),) +
+                            args[3:]
+                        )
+                    elif node.name == "currentdate":
+                        args = (
+                            args[:3] +
+                            (":not{}".format(args[3][1:]),) +
+                            args[4:]
+                        )
+                    elif node.name == "exists":
+                        args = ("not{}".format(args[0]),) + args[1:]
+                    negate = False
+                conditions.append(args)
+        return conditions
+
+    def get_filter_actions(self, name):
+        """Retrieve actions of the given filter."""
+        flt = self.getfilter(name)
+        if not flt:
+            return None
+        actions = []
+        for node in flt.walk():
+            if isinstance(node, commands.ActionCommand):
+                actions.append(node.args_as_tuple())
+        return actions
+
     def removefilter(self, name):
         """Remove a specific filter
 
@@ -355,8 +471,8 @@ class FiltersSet(object):
         :return: True if filter was disabled, False otherwise
         """
         name = self._unicode_filter_name(name)
-        ifcontrol = get_command_instance("if")
-        falsecmd = get_command_instance("false", ifcontrol)
+        ifcontrol = commands.get_command_instance("if")
+        falsecmd = commands.get_command_instance("false", ifcontrol)
         ifcontrol.check_next_arg("test", falsecmd)
         for f in self.filters:
             if f["name"] != name:
